/*
 * MaxDecayTaskAdvocate.hh
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file MaxDecayTaskAdvocate.hh
 *
 * Header file for the MaxDecayTaskAdvocate class.
 */

#ifndef _max_decay_task_advocate_hh
#define _max_decay_task_advocate_hh

#include <RealTimeTaskDelegateImpl.hh>

/**
 * An advocate that reacts quickly to increases in CPU usage, but slow to react
 * to decreases in usage, making it an ideal generic advocate.
 *
 * The actual implementation works by sending an advice that reflects the
 * highest usage seen in the last N periods, where N can be set using the
 * #SetDelegateAttribute method.
 */
class MaxDecayTaskAdvocate : public virtual RealTimeTaskDelegateImpl
{

public:
    
    /** Default constructor. */
    MaxDecayTaskAdvocate(void);

    /** Destructor. */
    virtual ~MaxDecayTaskAdvocate(void);
    
    /**
     * @copydoc BrokerDelegates::Delegate::SetDelegateAttribute
     *
     * Currently supported 'id' values are:
     *
     * @li @e decay-after The number of periods to wait before dropping the CPU
     * usage.
     */
    virtual void SetDelegateAttribute(const char *id,
				      const CORBA::Any &value)
	throw (CORBA::SystemException);

    /**
     * @copydoc BrokerDelegates::Delegate::GetDelegateAttribute
     *
     * Currently supported 'id' values are:
     *
     * @li @e decay-after The number of periods to wait before dropping the CPU
     * usage.
     * @li @e last-advice The last advice given by the advocate.
     */
    virtual CORBA::Any *GetDelegateAttribute(const char *id)
	throw (CORBA::SystemException);
    
    /** @copydoc Broker::RealTimeTask::PassCPU */
    virtual Broker::CPUReserve PassCPU(Broker::RealTimeTask_ptr rtt,
				       const Broker::CPUReserve &status,
				       const Broker::CPUReserve &advice,
				       const Broker::KeyedReportParameters &krp)
	throw (CORBA::SystemException);

private:

    /**
     * The last advice reported by this advocate.
     */
    CORBA::ULong md_LastAdvice;

    /**
     * The maximum observed usage that is still below the last advice given.
     */
    CORBA::ULong md_SubMax;

    /**
     * The number of periods left before the reservation will be dropped to the
     * value of md_SubMax.
     */
    unsigned int md_Decay;

    /**
     * The number of periods after which the advice should decay.
     */
    CORBA::ULong md_DecayAfter;
    
};

#endif
