/*
 * LoggingAdvocate.cc
 *
 * Copyright (c) 2003, 2004 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file LoggingAdvocate.cc
 *
 * Implementation of the LoggingAdvocate class.
 */

#include "config.h"

#include <string.h>
#include <sys/time.h>

#include <iostream>
#include <fstream>
#include <iomanip>

#include <time_util.h>

#include "LoggingAdvocate.hh"

using namespace std;

LoggingAdvocate::LoggingAdvocate(void)
{
    gettimeofday(&this->la_StartTime, NULL);
    this->la_LastCPULog = this->la_StartTime;
    this->la_CPULog = NULL;
    this->la_LastAdviceLog = this->la_StartTime;
    this->la_AdviceLog = NULL;
    this->la_CompleteLog = NULL;
    this->la_HitLog = NULL;
    this->la_MissLog = NULL;
}

LoggingAdvocate::~LoggingAdvocate()
{
    if( this->la_CPULog != NULL )
	delete this->la_CPULog;
    if( this->la_AdviceLog != NULL )
	delete this->la_AdviceLog;
    if( this->la_CompleteLog != NULL )
	delete this->la_CompleteLog;
    if( this->la_HitLog != NULL )
	delete this->la_HitLog;
    if( this->la_MissLog != NULL )
	delete this->la_MissLog;
}

void LoggingAdvocate::SetDelegateAttribute(const char *name,
					   const CORBA::Any &value)
    throw (CORBA::SystemException)
{
    if( (strcasecmp(name, "cpu_log") == 0) ||
	(strcasecmp(name, "cpu-log") == 0) ||
	(strcasecmp(name, "cpu log") == 0) )
    {
	const char *cpu_log;
	
	if( this->la_CPULog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("cpu log");
	}
	else if( value >>= cpu_log )
	{
	    this->la_CPULog = new ofstream(cpu_log);
	    (*this->la_CPULog) << setfill('0');
	    (*this->la_CPULog)
		<< "# start: "
		<< (long)this->la_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->la_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else if( (strcasecmp(name, "advice_log") == 0) ||
	     (strcasecmp(name, "advice-log") == 0) ||
	     (strcasecmp(name, "advice log") == 0) )
    {
	const char *advice_log;
	
	if( this->la_AdviceLog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("advice log");
	}
	else if( value >>= advice_log )
	{
	    this->la_AdviceLog = new ofstream(advice_log);
	    (*this->la_AdviceLog) << setfill('0');
	    (*this->la_AdviceLog)
		<< "# start: "
		<< (long)this->la_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->la_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else if( (strcasecmp(name, "complete_log") == 0) ||
	     (strcasecmp(name, "complete-log") == 0) ||
	     (strcasecmp(name, "complete log") == 0) )
    {
	const char *complete_log;
	
	if( this->la_CompleteLog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("complete log");
	}
	else if( value >>= complete_log )
	{
	    this->la_CompleteLog = new ofstream(complete_log);
	    (*this->la_CompleteLog) << setfill('0');
	    (*this->la_CompleteLog)
		<< "# start: "
		<< (long)this->la_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->la_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else if( (strcasecmp(name, "hit_log") == 0) ||
	     (strcasecmp(name, "hit-log") == 0) ||
	     (strcasecmp(name, "hit log") == 0) )
    {
	const char *hit_log;
	
	if( this->la_HitLog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("hit log");
	}
	else if( value >>= hit_log )
	{
	    this->la_HitLog = new ofstream(hit_log);
	    (*this->la_HitLog) << setfill('0');
	    (*this->la_HitLog)
		<< "# start: "
		<< (long)this->la_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->la_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else if( (strcasecmp(name, "miss_log") == 0) ||
	     (strcasecmp(name, "miss-log") == 0) ||
	     (strcasecmp(name, "miss log") == 0) )
    {
	const char *miss_log;
	
	if( this->la_MissLog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("miss log");
	}
	else if( value >>= miss_log )
	{
	    this->la_MissLog = new ofstream(miss_log);
	    (*this->la_MissLog) << setfill('0');
	    (*this->la_MissLog)
		<< "# start: "
		<< (long)this->la_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->la_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else
    {
	this->RealTimeTaskDelegateImpl::SetDelegateAttribute(name, value);
    }
}

CORBA::Any_ptr LoggingAdvocate::GetDelegateAttribute(const char *name)
    throw (CORBA::SystemException)
{
    CORBA::Any_var retval;
    
    if( name == NULL )
    {
	throw CORBA::BAD_PARAM();
    }
    else if( (strcasecmp(name, "last_advice") == 0) ||
	     (strcasecmp(name, "last-advice") == 0) ||
	     (strcasecmp(name, "last advice") == 0) )
    {
	retval = new CORBA::Any();
	(*retval) <<= this->la_LastAdvice;
    }
    else
    {
	retval = this->RealTimeTaskDelegateImpl::GetDelegateAttribute(name);
    }
    return( retval._retn() );
}

Broker::CPUReserve LoggingAdvocate::PassCPU(Broker::RealTimeTask_ptr rtt,
					    const Broker::CPUReserve &status,
					    const Broker::CPUReserve &advice,
					    const Broker::KeyedReportParameters &krp)
    throw (CORBA::SystemException)
{
    if( CORBA::is_nil(this->dm_RemoteObject.in()) )
    {
	throw CORBA::BAD_INV_ORDER();
    }

    this->la_LastAdvice = advice.Compute;
    
    if( this->la_CompleteLog != NULL )
    {
	struct timeval tv, diff;
	
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->la_StartTime, &diff);
	(*this->la_CompleteLog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " -0.01"
	    << endl;
    }
    if( this->la_HitLog != NULL )
    {
	struct timeval tv, st, diff;
	unsigned long long usecs;
	
	st.tv_sec = status.Compute / 1000000;
	st.tv_usec = status.Compute % 1000000;
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->la_LastLog, &diff);
	usecs = diff.tv_sec * 1000000;
	usecs += diff.tv_usec;
	if( usecs <= advice.Period )
	{
	    timersub(&tv, &this->la_StartTime, &diff);
	    (*this->la_HitLog)
		<< (long)diff.tv_sec
		<< "."
		<< setw(6) << (long)diff.tv_usec
		<< " "
		<< (long)st.tv_sec
		<< "."
		<< setw(6) << (long)st.tv_usec
		<< endl;
	}
	else
	{
	    // cerr << "miss " << usecs << " " << advice.Period << endl;
	}
    }
    if( this->la_MissLog != NULL )
    {
	struct timeval tv, st, diff;
	unsigned long long usecs;
	
	st.tv_sec = status.Compute / 1000000;
	st.tv_usec = status.Compute % 1000000;
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->la_LastLog, &diff);
	usecs = diff.tv_sec * 1000000;
	usecs += diff.tv_usec;
	if( usecs > advice.Period )
	{
	    timersub(&tv, &this->la_StartTime, &diff);
	    (*this->la_MissLog)
		<< (long)diff.tv_sec
		<< "."
		<< setw(6) << (long)diff.tv_usec
		<< " "
		<< (long)st.tv_sec
		<< "."
		<< setw(6) << (long)st.tv_usec
		<< endl;
	}
    }
    if( this->la_CPULog != NULL )
    {
	struct timeval tv, st, diff;
	
	st.tv_sec = status.Compute / 1000000;
	st.tv_usec = status.Compute % 1000000;
	timersub(&this->la_LastCPULog, &this->la_StartTime, &diff);
	(*this->la_CPULog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
	gettimeofday(&tv, NULL);
#if 0
	timersub(&tv, &this->la_StartTime, &diff);
	(*this->la_CPULog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
#endif
	this->la_LastCPULog = tv;
    }
    if( this->la_AdviceLog != NULL )
    {
	struct timeval tv, st, diff;

	st.tv_sec = advice.Compute / 1000000;
	st.tv_usec = advice.Compute % 1000000;
#if 0
	timersub(&this->la_LastAdviceLog, &this->la_StartTime, &diff);
	(*this->la_AdviceLog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
#endif
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->la_StartTime, &diff);
	(*this->la_AdviceLog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
	this->la_LastAdviceLog = tv;
    }
    gettimeofday(&this->la_LastLog, NULL);
    return this->RealTimeTaskDelegateImpl::PassCPU(rtt, status, advice, krp);
}
